/*
 *
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pvrusb2.h"
#include "pvrusb2-tuner.h"
#include "ivtv.h"

#ifndef I2C_DRIVERID_SAA7115
// Using temporary hack for missing I2C driver-ID for saa7115
#define I2C_DRIVERID_SAA7115 I2C_DRIVERID_EXP1
#endif

/*

  This module attempts to implement a compliant I2C adapter for the
  pvrusb2 device.  By doing this we can then make use of existing
  functionality in V4L (e.g. tuner.c) rather than rolling our own.

 */

static unsigned int i2c_scan = 0;
module_param(i2c_scan, int, 0444);
MODULE_PARM_DESC(i2c_scan,"scan i2c bus at insmod time");

static int pvr_i2c_write(struct pvr *pvr,  /* Context */
			 u8 i2c_addr,      /* I2C address we're talking to */
			 u8 *data,         /* Data to write */
			 u16 length)       /* Size of data to write */
{
    /* Return value - default 0 means success */
    int ret;

//    pvr_debug("pvr_i2c_write");

    if (!data) length = 0;
    if (length > (sizeof(pvr->cmd_buffer) - 3)) return -ENOTSUPP;

    down(&pvr->ctl_sem);

    /* Clear the command buffer (likely to be paranoia) */
    memset(pvr->cmd_buffer, 0, sizeof(pvr->cmd_buffer));

    /* Set up command buffer for an I2C write */
    pvr->cmd_buffer[0] = 0x08;      /* write prefix */
    pvr->cmd_buffer[1] = i2c_addr;  /* i2c addr of chip */
    pvr->cmd_buffer[2] = length;    /* length of what follows */
    if (length) memcpy(pvr->cmd_buffer + 3, data, length);

    /* Do the operation */
    ret = pvr_send_request(pvr,
			   pvr->cmd_buffer,
			   length + 3,
			   pvr->cmd_buffer,
			   1);
    if (!ret) {
	if (pvr->cmd_buffer[0] != 8) {
	    ret = -EIO;
	    if (pvr->cmd_buffer[0] != 7) {
		pvr_debug("pvrusb2: unexpected status from i2_write[%d]: %d",
			  i2c_addr,pvr->cmd_buffer[0]);
	    }
	}
    }
#ifdef notdef
    pvr_debug("pvrusb2 i2c_write(%d) ret=%d stat=%d",i2c_addr,ret,
	      pvr->cmd_buffer[0]);
#endif

    up(&pvr->ctl_sem);

    return ret;
}

static int pvr_i2c_read(struct pvr *pvr,   /* Context */
			u8 i2c_addr,       /* I2C address we're talking to */
			u8 *data,          /* Data to write */
			u16 dlen,          /* Size of data to write */
			u8 *res,           /* Where to put data we read */
			u16 rlen)          /* Amount of data to read */
{
    /* Return value - default 0 means success */
    int ret;

//    pvr_debug("pvr_i2c_read");

    if (!data) dlen = 0;
    if (dlen > (sizeof(pvr->cmd_buffer) - 4)) return -ENOTSUPP;
    if (res && (rlen > (sizeof(pvr->cmd_buffer) - 1))) return -ENOTSUPP;

    down(&pvr->ctl_sem);

    /* Clear the command buffer (likely to be paranoia) */
    memset(pvr->cmd_buffer, 0, sizeof(pvr->cmd_buffer));

    /* Set up command buffer for an I2C write followed by a read */
    pvr->cmd_buffer[0] = 0x09;  /* read prefix */
    pvr->cmd_buffer[1] = dlen;  /* arg length */
    pvr->cmd_buffer[2] = rlen;  /* answer length. Device will send one
				   more byte (status). */
    pvr->cmd_buffer[3] = i2c_addr;  /* i2c addr of chip */
    if (dlen) memcpy(pvr->cmd_buffer + 4, data, dlen);

    /* Do the operation */
    ret = pvr_send_request(pvr,
			   pvr->cmd_buffer,
			   4 + dlen,
			   pvr->cmd_buffer,
			   rlen + 1);
    if (!ret) {
	if (pvr->cmd_buffer[0] != 8) {
	    ret = -EIO;
	    if (pvr->cmd_buffer[0] != 7) {
		pvr_debug("pvrusb2: unexpected status from i2_read[%d]: %d",
			  i2c_addr,pvr->cmd_buffer[0]);
	    }
	}
    }

#ifdef notdef
    pvr_debug("pvrusb2 i2c_read(%d) ret=%d stat=%d",i2c_addr,ret,pvr->cmd_buffer[0]);
#endif
    /* Copy back the result */
    if (res && rlen) {
	if (ret) {
	    /* Error, just blank out the return buffer */
	    memset(res, 0, rlen);
	} else {
	    memcpy(res, pvr->cmd_buffer + 1, rlen);
	}
    }

    up(&pvr->ctl_sem);

    return ret;
}

/* This is a very, very limited I2C adapter implementation.  We can
   only support what we actually know will work on the device... */
static int pvr_i2c_xfer(struct i2c_adapter *i2c_adap,
			struct i2c_msg msgs[],
			int num)
{
    int ret = -ENOTSUPP;
    struct pvr *pvr = (struct pvr *)(i2c_adap->algo_data);

    if ((msgs[0].flags & I2C_M_NOSTART)) {
	pvr_debug("pvrusb2: i2c refusing I2C_M_NOSTART");
	goto done;
    }

    if (num == 1) {
	if (msgs[0].flags & I2C_M_RD) {
	    /* Simple read */
	    ret = 1;
	    if (pvr_i2c_read(pvr,msgs[0].addr,
			     0,0,
			     msgs[0].buf,msgs[0].len)) {
		ret = -EIO;
	    }
	    goto done;
	} else {
	    /* Simple write */
	    ret = 1;
	    if (pvr_i2c_write(pvr,msgs[0].addr,msgs[0].buf,msgs[0].len)) {
		ret = -EIO;
	    }
	    goto done;
	}
    } else if (num == 2) {
	if ((!((msgs[0].flags & I2C_M_RD))) && (msgs[1].flags & I2C_M_RD)) {
	    /* Write followed by atomic read */
	    ret = 2;
	    if (pvr_i2c_read(pvr,msgs[0].addr,
			     msgs[0].buf,msgs[0].len,
			     msgs[1].buf,msgs[1].len)) {
		ret = -EIO;
	    }
	    goto done;
	} else {
	    pvr_debug("pvrusb2: i2c refusing complex transfer"
		      " read0=%d read1=%d",
		      (msgs[0].flags & I2C_M_RD),
		      (msgs[1].flags & I2C_M_RD));
	}
    } else {
	pvr_debug("pvrusb2: i2c refusing %d phase transfer",num);
    }

 done:
    return ret;
}

static int pvr_i2c_control(struct i2c_adapter *adapter,
			   unsigned int cmd, unsigned long arg)
{
    return 0;
}

static u32 pvr_i2c_functionality(struct i2c_adapter *adap)
{
    return I2C_FUNC_SMBUS_EMUL;
}

static int pvr_i2c_attach_inform(struct i2c_client *client)
{
    struct pvr *pvr = (struct pvr *)(client->adapter->algo_data);
    int id;
    id = client->driver->id;
    pvr_debug("%s: i2c_attach [client=%s @ 0x%x id=0x%x]",
	      pvr->name,i2c_clientname(client),
	      client->addr,id);
    if (!(pvr->i2c_msp3400_client) && (id == I2C_DRIVERID_MSP3400)) {
	pvr_debug("%s: attaching msp3400 I2C client",
		  pvr->name);
	pvr->i2c_msp3400_client = client;
    }
    if (!(pvr->i2c_tuner_client) && (id == I2C_DRIVERID_TUNER)) {
	pvr_debug("%s: attaching tuner I2C client",
		  pvr->name);
	pvr->i2c_tuner_client = client;
	pvr_tuner_set_type(pvr);
    }
    if (!(pvr->i2c_saa7115_client) && (id == I2C_DRIVERID_SAA7115)) {
	pvr_debug("%s: attaching saa7115 I2C client",
		  pvr->name);
	pvr->i2c_saa7115_client = client;
    }

    return 0;
}

int pvr_i2c_cmd(struct i2c_client *cp,unsigned int cmd,void *arg)
{
    int stat;
    if (!cp) return -EINVAL;
    if (!(cp->driver)) return -EINVAL;
    if (!(cp->driver->command)) return -EINVAL;
    if (!try_module_get(cp->driver->owner)) return -EAGAIN;
    stat = cp->driver->command(cp,cmd,arg);
    module_put(cp->driver->owner);
    return stat;
}

int pvr_i2c_tuner_cmd(struct pvr *pvr,unsigned int cmd,void *arg)
{
    int stat = pvr_i2c_cmd(pvr->i2c_tuner_client,cmd,arg);
    if (stat < 0) pvr_debug("pvr_i2c_tuner_cmd failed with status %d",stat);
    return stat;
}

int pvr_i2c_msp3400_cmd(struct pvr *pvr,unsigned int cmd,void *arg)
{
    int stat = pvr_i2c_cmd(pvr->i2c_msp3400_client,cmd,arg);
    if (stat < 0) pvr_debug("pvr_i2c_msp3400_cmd failed with status %d",stat);
    return stat;
}

int pvr_i2c_saa7115_cmd(struct pvr *pvr,unsigned int cmd,void *arg)
{
    int stat = pvr_i2c_cmd(pvr->i2c_saa7115_client,cmd,arg);
    if (stat < 0) pvr_debug("pvr_i2c_saa7115_cmd failed with status %d",stat);
    return stat;
}

static int pvr_i2c_detach_inform(struct i2c_client *client)
{
    struct pvr *pvr = (struct pvr *)(client->adapter->algo_data);
    pvr_debug("%s: pvr_i2c_detach [client=%s @ 0x%x]",
	      pvr->name,i2c_clientname(client),
	      client->addr);
    if (pvr->i2c_msp3400_client == client) {
	pvr_debug("%s: detaching msp3400 I2C client",
		  pvr->name);
	pvr->i2c_msp3400_client = 0;
    }
    if (pvr->i2c_tuner_client == client) {
	pvr_debug("%s: detaching tuner I2C client",
		  pvr->name);
	pvr->i2c_tuner_client = 0;
    }
    if (pvr->i2c_saa7115_client == client) {
	pvr_debug("%s: detaching saa7115 I2C client",
		  pvr->name);
	pvr->i2c_saa7115_client = 0;
    }

    return 0;
}

static struct i2c_algorithm pvr_i2c_algo_template = {
    .id            = I2C_ALGO_BIT | I2C_HW_B_BT848,
    .master_xfer   = pvr_i2c_xfer,
    .algo_control  = pvr_i2c_control,
    .functionality = pvr_i2c_functionality,
};

static struct i2c_adapter pvr_i2c_adap_template = {
    .owner         = THIS_MODULE,
    .class	   = I2C_CLASS_TV_ANALOG,
    .id            = I2C_ALGO_BIT | I2C_HW_B_BT848,
    .client_register = pvr_i2c_attach_inform,
    .client_unregister = pvr_i2c_detach_inform,
};

static void do_i2c_scan(struct pvr *pvr)
{
    struct i2c_msg msg[1];
    int i,rc;
    msg[0].addr = 0;
    msg[0].flags = I2C_M_RD;
    msg[0].len = 0;
    msg[0].buf = 0;
    for (i = 0; i < 128; i++) {
	msg[0].addr = i;
	rc = i2c_transfer(&pvr->i2c_adap,msg,sizeof(msg)/sizeof(msg[0]));
	if (rc < 0) continue;
	printk("%s: i2c scan: found device @ 0x%x\n",pvr->name,i);
    }
}

void pvr_i2c_init(struct pvr *pvr)
{
    memcpy(&pvr->i2c_adap,&pvr_i2c_adap_template,sizeof(pvr->i2c_adap));
    memcpy(&pvr->i2c_algo,&pvr_i2c_algo_template,sizeof(pvr->i2c_algo));
    strlcpy(pvr->i2c_adap.name,pvr->name,sizeof(pvr->i2c_adap.name));
    strlcpy(pvr->i2c_algo.name,pvr->name,sizeof(pvr->i2c_algo.name));
    pvr->i2c_adap.algo = &pvr->i2c_algo;
    pvr->i2c_adap.algo_data = pvr;
    i2c_add_adapter(&pvr->i2c_adap);
    pvr->i2c_linked = !0;
    if (i2c_scan) do_i2c_scan(pvr);
}

void pvr_i2c_done(struct pvr *pvr)
{
    if (pvr->i2c_linked) {
	i2c_del_adapter(&pvr->i2c_adap);
	pvr->i2c_linked = 0;
    }
}

