/*
 *  
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef __PVRUSB2_H
#define __PVRUSB2_H


#include <linux/videodev.h>
#include <linux/usb.h>
#include <linux/i2c.h>
#include <media/tveeprom.h>

/* size of a firmware chunk */
#define FIRMWARE_CHUNK_SIZE 0x2000

/* Various files we will load with firmware_entry */
#define FIRMWARE1_FILE  "pvrusb2.f1"
#define FIRMWARE2_FILE  "pvrusb2.f2"

/* Buffer management constants */
#define PVR_VID_BUFFSIZE        PAGE_ALIGN(0x4000)
#define PVR_CTL_BUFFSIZE        0x0100
#define PVR_NBUFF               32       /* Max number of video buffers */

#define PVR_VID_ENDPOINT        0x84
#define PVR_UNK_ENDPOINT        0x86    /* maybe raw yuv ? */
#define PVR_VBI_ENDPOINT        0x88
#define PVR_CTL_WRITE_ENDPOINT  0x01
#define PVR_CTL_READ_ENDPOINT   0x81
#define PVR_FIRMWARE_ENDPOINT   0x02

#define PVR_DEBUG 1
#define pvr_debug(...)    do { if(PVR_DEBUG) info(__VA_ARGS__); } while(0)
/* BUG_ON is a bit aggressive... maybe find a replacement */
#define PVR_BUG_ON(cond)   BUG_ON(cond)

#define VID_HARDWARE_PVRUSB2    38  /* FIXME : need a good value */

#define PVR_NSTREAM 4

#define PVR_DECOMPOSE_LE(t,i,d) \
    do {    \
        (t)[i] = (d) & 0xff;\
        (t)[i+1] = ((d) >> 8) & 0xff;\
        (t)[i+2] = ((d) >> 16) & 0xff;\
        (t)[i+3] = ((d) >> 24) & 0xff;\
    } while(0)

#define PVR_DECOMPOSE_BE(t,i,d) \
    do {    \
        (t)[i+3] = (d) & 0xff;\
        (t)[i+2] = ((d) >> 8) & 0xff;\
        (t)[i+1] = ((d) >> 16) & 0xff;\
        (t)[i] = ((d) >> 24) & 0xff;\
    } while(0)

enum pvr_device_state {
    _unstable=0,        /* device is fucked up. Absorbant state. */
    _reset,
    _need_firmware,     /* state 0 */
    _need_config,       /* encoder firmware in. Need to configure capture. */
    _stopped,           /* Device is ready to start capture. */
    _streaming
};

enum pvr_memory_state {
    _freed,
    _allocated,
    _mapped,    /* at least one buffer is mapped */
};

enum pvr_io_method {
    _none,
    _read,
    _mmap,
    _uptr
};

enum pvr_config {
    _empty,
    _mpeg,
    _vbi,
    _radio,
};

struct pvr_cmd {
    u8 length; 
    u8 ret;
    u8 *cmd;
};

#define PVR_SIZEOf_CMD(c)   (2*sizeof(short)+sizeof(char)*(c)->length)
#define PVR_CMD(r,c)   { .length = sizeof(c)-1, .ret = r , .cmd = c }
#define PVR_DECLARE_CMD(name,r,c) struct pvr_cmd name = PVR_CMD(r,c)

struct pvr_mapping {
	struct pvr_stream *stream;
	int count;
};

struct pvr_buffer {
    struct pvr_stream *stream;
    void *buff;     /* Allocated buffer. Page aligned one is in purb */
    u8 *data;     /* First unread data byte */
    size_t size;    /* Size of unread data (for read-io) */
    struct urb *purb;
    short submitted;
    struct pvr_mapping *map;    /* NULL if not mapped */
    int index;
};

struct pvr_uptr {
    unsigned long ptr;
    unsigned int length;
};

struct pvr_stream {
    struct pvr *pvr;
    struct video_device *vdev;
    int v4l_type;

    struct semaphore sem;
    int user;

    unsigned long last_buffer_timestamp;
    enum pvr_memory_state mem_state;
    enum pvr_io_method io_method;
    struct pvr_fh *io_owner;
    enum pvr_config config;

    int endpoint;   /* usb endpoint to read */
    atomic_t submitted_urbs;
    wait_queue_head_t urbs_collected;
    wait_queue_head_t wait_data;

    struct pvr_buffer buffer[PVR_NBUFF];        /* allocated buffers */
    int buffer_count;   /* how many buffer do we want ? */
    int map_count;      /* number of mapped buffers */
    struct pvr_buffer *rx_buffer[PVR_NBUFF];    /* ring of recv buffers */
    spinlock_t ring_lock;   /* we don't want simultaneous add to the ring. */
    unsigned int rx1;
    /* might want to use atomic for smp ?*/
    unsigned int volatile rx2;
    /* ring for user pointers (acces under sem lock)*/
    struct pvr_uptr userptr[PVR_NBUFF];
    unsigned int uptr1;
    unsigned int uptr2;
    int align;
};

struct pvr_fh {
    struct pvr_stream *stream;
	enum v4l2_priority prio;
};

struct pvr_video_config {
    int input;
    int standard;
    int width;
    int height_full;
    int bitrate;
    int peakbitrate;
    int vbr;
    int brightness;
    int contrast;
    int saturation;
    int hue;

};

struct pvr_audio_config {
    int input;
    int bitrate;
    int srate;
    int layer;    /* only layer 2 seems supported */
    int crc;
    int emphasis;
    int volume;
    int muted;
    int balance;
    int bass;
    int treble;
    int lang;
    int mode;
    int status_bilingual;
    int status_stereo;
};


struct pvr {
    struct usb_device* udev;
    char name[32];
    struct semaphore sem;
    int nr;
    int disconnecting;
    int renum;
    enum pvr_device_state dev_state;
    int supported_std;
    struct v4l2_prio_state prio;

    /* Tuner / frequency control stuff */
    struct v4l2_frequency freq;
    struct tveeprom tvdata;

    /* I2C stuff */
    struct i2c_adapter i2c_adap;
    struct i2c_algorithm i2c_algo;
    int i2c_linked;
    struct i2c_client *i2c_tuner_client;
    struct i2c_client *i2c_msp3400_client;
    struct i2c_client *i2c_saa7115_client;

    /* streams */
    struct pvr_stream* stream[PVR_NSTREAM];
    int stream_count;

    /* our control pool */
    struct semaphore ctl_sem;
    unsigned char cmd_buffer[PVR_CTL_BUFFSIZE];  /* used under ctl_sem lock */

#ifdef notdef
    /* remote stuff */
    struct input_dev* idev;
    struct timer_list remote_timer;    /* will read key pressed on the remote. */
    struct work_struct remote_task;
    struct semaphore task_sem;
    int last_key;
#endif

    /* audio/video parameters */
    enum pvr_config config;

    struct pvr_video_config video;
    struct pvr_audio_config audio;
};

int pvr_send_sequence(struct pvr *, struct pvr_cmd *);
int pvr_send_request(struct pvr *, void *, unsigned int, void *, unsigned int);
int pvr_do_command(struct pvr *, struct pvr_cmd *);

int pvr_write_register(struct pvr *, u16, u32);
int pvr_read_register(struct pvr *, u16, u32 *);
int pvr_write_u16(struct pvr *, u16, int);
int pvr_write_u8(struct pvr *, u8, int);

void pvr_reconfigure(struct pvr *);
void pvr_state_change(struct pvr *, enum pvr_device_state);
void pvr_state_above(struct pvr *, enum pvr_device_state);
void pvr_state_below(struct pvr *, enum pvr_device_state);

void pvr_i2c_init(struct pvr *);
void pvr_i2c_done(struct pvr *);
int pvr_i2c_tuner_cmd(struct pvr *,unsigned int,void *);
int pvr_i2c_msp3400_cmd(struct pvr *,unsigned int,void *);
int pvr_i2c_saa7115_cmd(struct pvr *,unsigned int,void *);

int pvr_config_claim(struct pvr *, struct pvr_fh *);
int pvr_config_release(struct pvr *, struct pvr_fh *);
int pvr_config_apply(struct pvr *, enum pvr_config);

#endif /* __PVRUSB2_H */
